#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#ifdef _WIN32
#include <winmmap.h>
#else
#include <sys/mman.h>
#endif
#include <fcntl.h>
#include <string.h>
#include <csv.h>
#include <section.h>

/*
 * Extract sections from a CSV file
 */ 
int readSectionsFromCSV(char* iFileName, char iSeparator, Section** iSection, size_t* iSectionCount) {
  int fd, i, found, line, ok;
  struct stat sb;
  char *data;
  Section* ptr;
  CSVState state;
  CSVToken token[4];

  /* Set token types */
  token[0].type = STRING;
  token[1].type = STRING;
  token[2].type = STRING;
  token[3].type = STRING;	

  for(i=0; i<4; ++i) {
  	token[i].value.string = NULL;
  }

  ok = 1;

  /* Open file */
  if((fd = open(iFileName, O_RDONLY)) < 0) {
  	fprintf(stderr, "Error: Unable to open %s.\n", iFileName);
  	return 0;
  }

  /* Get infos */
  if(fstat(fd, &sb) < 0) {
  	fprintf(stderr, "Error: Could not stat %s.\n",iFileName);
  	close(fd);
  	return 0;
  }

  if(sb.st_size < 4) {
  	fprintf(stderr, "Error: %s is too small.\n",iFileName);
  	close(fd);
  	return 0;
  }

  /* Mmap it */
  data = (char*)mmap(NULL, sb.st_size, PROT_READ, MAP_SHARED, fd, 0);
  if(data == MAP_FAILED) {
  	fprintf(stderr, "Error: Could not mmap %s.\n",iFileName);
  	close(fd);
  	return 0;
  }
	
  /* Initalize csv state */
  CSVSetState(data, data + sb.st_size, &state);
  
  /* Count lines */
  for(; CSVJumpToNextLine(&state); ++*iSectionCount);
  ++*iSectionCount;

  /* Allocate sctions */
  *iSection = (Section*)malloc(*iSectionCount * sizeof(Section));
  
  /* Reset state */
  CSVSetState(data, data + sb.st_size, &state);
  
  line = 0;
  ptr = *iSection + 5;
  do {
  	/* Extract tokens */
	found = CSVExtractTokensFromLine(&state, iSeparator, token, 4);

	/* Not enough tokens found */
	if(found < 4) {
		fprintf(stderr, "Error : too few tokens on line %d\n", line);
		ok = 0;
	} 
	else {
		/* Process tokens :			*/
		/*		1. type				*/
		if(strcmp(token[0].value.string, "code") == 0) {
			ptr[line].type = CODE;
		}
		else {
			/* by default a section is made of data */
			ptr[line].type = DATA;
		}
		/*		2. name				*/
		ptr[line].name = strdup(token[1].value.string);
		/*		3. start address	*/
		ptr[line].start = strtol(token[2].value.string, (char**)NULL, 16);
		/*		3. end address		*/
		ptr[line].end = strtol(token[3].value.string, (char**)NULL, 16);
		
		if(ptr[line].end < ptr[line].start) {
			fprintf(stderr, "Error: end offset is smaller than start offset.\n");
			fprintf(stderr, "Error: start=%0lx end=%0lx\n", ptr[line].start, ptr[line].end);
			ok = 0;
		}
	}

	/* Cleanup strings */
	for(i=0; i<found; ++i) {
		if((token[i].type == STRING) &&
		   (token[i].value.string != NULL) ) {
			free(token[i].value.string);
			token[i].value.string = NULL;
		}
	}
	
	++line;
  }while(CSVJumpToNextLine(&state) && ok);
	
  /* Release map */
  if(munmap(data, sb.st_size) < 0) {
    fprintf(stderr, "Error: Could not mmap %s.\n",iFileName);
  	close(fd);
  	return 0;
  }

  /* Close file */
  close(fd);

  return ok;
}
